/**
 * Created by Ludo on 20-12-2015.
 */

function array2dcopy(array) {
    var ret = [];
    for (var i = 0; i < array.length; i++) {
        ret.push(array[i].slice());
    }
    return ret;
}

/*******************************************************************************
 * LessModel is the model of the Less visualisation.
 * This part makes sure that the board transitions can be made.
 * When the board is changed, it sends the board to all its controllers.
 ******************************************************************************/

function LessModel() {
    this.char1 = '1'.charCodeAt(0);
    this.charA = 'a'.charCodeAt(0);

    this.size = 8;
    this.curMove = 0;
    this.curPhase = 0;
    this.controllers = [];

    this.startPosition = [];
    for (var i = this.size; i--;) {
        var arr = [];
        for (var j = this.size; j--;) arr.push(-1);
        this.startPosition.push(arr);
    }

    var positions = [
        [
            {row: this.size - 1, col: 0},
            {row: this.size - 1, col: 1},
            {row: this.size - 2, col: 0},
            {row: this.size - 2, col: 1}
        ], // bottom left
        [
            {row: 0, col: 0},
            {row: 0, col: 1},
            {row: 1, col: 0},
            {row: 1, col: 1}
        ], // top left
        [
            {row: this.size - 1, col: this.size - 1},
            {row: this.size - 1, col: this.size - 2},
            {row: this.size - 2, col: this.size - 1},
            {row: this.size - 2, col: this.size - 2}
        ], // bottom right
        [
            {row: 0, col: this.size - 1},
            {row: 0, col: this.size - 2},
            {row: 1, col: this.size - 1},
            {row: 1, col: this.size - 2}
        ] // top right
    ];
    for (var p = positions.length; p--;) {
        for (var stone = positions[p].length; stone--;) {
            this.startPosition[positions[p][stone].row][positions[p][stone].col] = p;
        }
    }

    this.curPosition = array2dcopy(this.startPosition);
}

LessModel.prototype.getSize = function () {
    return this.size;
};

LessModel.prototype.setWalls = function (walls) {
    this.walls = walls;
    return this;
};

LessModel.prototype.setMoves = function (moves) {
    if (moves[moves.length - 1] != 'Quit') moves.push('Quit');

    var i;
    this.moves = moves;
    this.endPosition = array2dcopy(this.startPosition);
    this.phaseStarts = [-1];
    for (i = 0; i < moves.length && moves[i] != 'Move' && moves[i] != 'Quit'; i++) {
        this.transition(this.endPosition, moves[i], true);
    }
    this.quitPosition = array2dcopy(this.endPosition);
    this.filterPlayer(this.quitPosition, -1);
    var curMovePosition = undefined, movesFound = false;
    for (; i < moves.length; i++) {
        if (moves[i] == 'Move' || moves[i] == 'Quit') {
            if (this.phaseStarts.length > 1)
                for (var row = 0; row < curMovePosition.length; row++) {
                    for (var col = 0; col < curMovePosition[row].length; col++) {
                        if (curMovePosition[row][col] >= 0) {
                            this.quitPosition[row][col] = curMovePosition[row][col];
                        }
                    }
                }
            if (moves[i] == 'Move') {
                curMovePosition = array2dcopy(this.endPosition);
                this.filterPlayer(curMovePosition, this.phaseStarts.length - 1);
                movesFound = true;
            }

            this.phaseStarts.push(i);
        } else if (moves[i] != 'Nil') {
            this.transition(curMovePosition, moves[i], true);
        }
    }
    if (!movesFound) {
        this.quitPosition = array2dcopy(this.endPosition);
    }
    return this;
};

LessModel.prototype.addController = function (controller) {
    this.controllers.push(controller);
    if (controller.setWalls) controller.setWalls(this.walls);
    if (controller.setMoves) controller.setMoves(this.moves);
    if (controller.onBoardChanged)
        controller.onBoardChanged(this.curMove, this.curPosition);
};

LessModel.prototype.boardChanged = function () {
    // fire board change
    for (var i = this.controllers.length; i--;) {
        if (this.controllers[i].onBoardChanged)
            this.controllers[i].onBoardChanged(this.curMove, this.curPosition);
    }
};

LessModel.prototype.transition = function (position, move, forward) {
    if (move == 'Nil') return;

    if (forward == undefined) forward = true;

    var move2 = move.split(':'), i, frow, fcol, trow, tcol;
    if (forward) {
        for (i = 0; i < move2.length; i++) {
            frow = move2[i].charCodeAt(0) - this.charA;
            fcol = move2[i].charCodeAt(1) - this.char1;
            trow = move2[i].charCodeAt(2) - this.charA;
            tcol = move2[i].charCodeAt(3) - this.char1;
            if (trow != frow || tcol != fcol) {
                position[trow][tcol] = position[frow][fcol];
                position[frow][fcol] = -1;
            }
        }
    } else {
        for (i = move2.length; i--;) {
            frow = move2[i].charCodeAt(0) - this.charA;
            fcol = move2[i].charCodeAt(1) - this.char1;
            trow = move2[i].charCodeAt(2) - this.charA;
            tcol = move2[i].charCodeAt(3) - this.char1;
            if (trow != frow || tcol != fcol) {
                position[frow][fcol] = position[trow][tcol];
                position[trow][tcol] = -1;
            }
        }
    }
};

LessModel.prototype.filterPlayer = function (position, player) {
    for (var row = 0; row < position.length; row++) {
        for (var col = 0; col < position[row].length; col++) {
            if (position[row][col] >= 0 && position[row][col] != player) {
                position[row][col] = -1;
            }
        }
    }
};

LessModel.prototype.getPhase = function (move) {
    var phase = this.phaseStarts.length - 1;
    while (phase > 0 && this.phaseStarts[phase] >= move) phase--;
    return phase;
};

LessModel.prototype.getMove = function () {
    return this.curMove;
};

// returns whether the move has been processed
LessModel.prototype.setMove = function (move) {
    if (move < 0 || move > this.moves.length || move == this.curMove) return move == this.curMove;

    var phase = this.getPhase(move);
    if (this.curPhase != phase) {
        // phase transition
        this.curPhase = phase;
        this.curMove = this.phaseStarts[phase] + 1;

        // edit curPosition
        if (phase == 0) {
            this.curPosition = array2dcopy(this.startPosition);
        } else if (phase == this.phaseStarts.length - 1) {
            this.curPosition = array2dcopy(this.quitPosition);
        } else {
            this.curPosition = array2dcopy(this.endPosition);
            this.filterPlayer(this.curPosition, phase - 1);
        }
    }

    // move backwards
    while (this.curMove > move) {
        this.transition(this.curPosition, this.moves[--this.curMove], false);
    }
    // move forwards
    while (this.curMove < move) {
        this.transition(this.curPosition, this.moves[this.curMove++], true);
    }

    this.boardChanged();
    return true;
};


/*******************************************************************************
 * LessController handles the interaction between the user and the LessModel
 ******************************************************************************/

function LessController(less, board, prev, next, playPause, moveTable) {
    this.colorWall = '#00aced';
    this.wallWidth = '5px';
    this.doubleWallWidth = '10px';

    this.stoneSize = 50;
    this.stonePadding = 2;
    this.colorStone = ['#ffff00', '#000000', '#ffffff', '#ff0000'];
    this.namesStone = ['Yellow', 'Black', 'White', 'Red'];

    this.less = less;

    this.board = document.getElementById(board);
    this.prev = document.getElementById(prev);
    this.next = document.getElementById(next);
    this.playPause = document.getElementById(playPause);
    this.moveTable = document.getElementById(moveTable);

    this.registerEvents();
}

LessController.prototype.setPlayerNames = function () {
    this.names = [];
    for (var i = 0; i < arguments.length; i++) {
        if ((typeof arguments[i]) == 'string') {
            this.names.push(arguments[i]);
        }
    }
    return this;
};

LessController.prototype.registerEvents = function () {
    var self = this, stopRunning = function () {
        self.isRunning = false;
        self.playPause.innerText = 'Play';
    };
    this.prev.onclick = function () {
        stopRunning();
        self.less.setMove(self.less.getMove() - 1);
    };
    this.next.onclick = function () {
        stopRunning();
        self.less.setMove(self.less.getMove() + 1);
    };
    this.playPause.onclick = function () {
        if (self.isRunning) {
            stopRunning();
        } else {
            self.isRunning = true;
            self.playPause.innerText = 'Pause';
            (function run() {
                if (self.isRunning) {
                    if (!self.less.setMove(self.less.getMove() + 1)) {
                        // last move has been reached.
                        stopRunning();
                    } else {
                        setTimeout(run, 1000); // 4 moves per second
                    }
                }
            })();
        }
    };
};

LessController.prototype._getCell = function (row, col) {
    return this.grid.children[row].children[col + 1]; // + 1 for header of this row
};

LessController.prototype.createTable = function () {
    var size = this.less.getSize(), i, j;
    var table = document.createElement('table'), tbody = document.createElement('tbody');
    var thead = document.createElement('thead'), theader = document.createElement('tr');
    for (i = 0; i <= size; i++) {
        var th = document.createElement('th');
        if (i > 0) th.innerText = '' + i;
        theader.appendChild(th);
    }
    thead.appendChild(theader);
    table.appendChild(thead);

    for (i = 0; i < size; i++) {
        var tr = document.createElement('tr');

        var colHeader = document.createElement('th');
        colHeader.innerText = String.fromCharCode(97 + i); // 'a'
        tr.appendChild(colHeader);

        for (j = 0; j < size; j++) tr.appendChild(document.createElement('td'));
        tbody.appendChild(tr);
    }

    this.grid = tbody;
    table.appendChild(tbody);
    while (this.board.firstChild)
        this.board.removeChild(this.board.firstChild);
    this.board.appendChild(table);
};

LessController.prototype.buildWalls = function (walls) {
    var wallRow = 0, wallCol = 0, wallHorizontal = true, size = this.less.getSize(), i;
    for (i = 0; i < walls.length && wallRow < size; i++) {
        // 0 | 1 | 2
        if (walls.charAt(i) == '1' || walls.charAt(i) == '2') {
            var double = (walls.charAt(i) == '2');
            var side = wallHorizontal ? 'Right' : 'Bottom';

            var element = this._getCell(wallRow, wallCol);
            element.style['border' + side + 'Color'] = this.colorWall;
            element.style['border' + side + 'Style'] = double ? 'double' : 'solid';
            element.style['border' + side + 'Width'] = double ? this.doubleWallWidth : this.wallWidth;
        }

        // calc next position
        wallCol++;
        if (wallHorizontal) {
            if (wallCol >= size - 1) {
                wallCol = 0;
                // same wallRow
                wallHorizontal = false;
            }
        } else {
            if (wallCol >= size) {
                wallCol = 0;
                wallRow++;
                wallHorizontal = true;
            }
        }
    }
    if (i != walls.length) {
        console.log('Walls length doesn\'t match board dimensions');
    }
};

LessController.prototype.buildMoveTable = function (moves) {
    var table = document.createElement('table'), thead = document.createElement('thead');
    var theader = document.createElement('tr'), self = this, i;
    for (i = 0; i < 5; i++) {
        var th = document.createElement('th');
        if (i > 0) {
            var playerName = this.namesStone[i - 1];
            if (this.names && i - 1 < this.names.length) {
                // cut-off on more than 20 characters in the name
                var name = this.names[i - 1];
                if (name.length > 20) name = name.slice(0, 17) + '...';
                playerName += ' (' + name + ')';
            }
            th.innerText = playerName;
        }
        theader.appendChild(th);
    }

    thead.appendChild(theader);
    table.appendChild(thead);

    var tbody = document.createElement('tbody');
    var player = -1, movesHad = 0;

    for (i = 0; i <= moves.length; i++) {
        var tr = document.createElement('tr');

        var moveNr = document.createElement('th');
        moveNr.innerText = (i + 0) + '.';
        if (i % 2 == 1) tr.setAttribute('class', 'zebra');
        tr.appendChild(moveNr);

        var currentMove = i == 0 ? 'Start' : moves[i - 1];
        if (i > 0) {
            if (currentMove == 'Move') {
                player = movesHad++;
            } else if (movesHad == 0) {
                player = (player + 1) % 4;
            }
        }
        for (var j = 0; j < 4; j++) {
            var td = document.createElement('td');
            if (i == 0) {
                if (j == 0) td.innerText = currentMove;
            } else if (currentMove == 'Quit' || player == j) {
                td.innerText = currentMove;
            }
            tr.appendChild(td);
        }


        tr.onclick = (function (move) {
            return function () {
                self.less.setMove(move);
            }
        })(i);

        tbody.appendChild(tr);
    }

    this.updateSelectedRow = function (moveNr) {
        var lastCurMove = document.getElementById('cur-move');
        if (lastCurMove) {
            lastCurMove.removeAttribute('id');
        }
        if (moveNr >= 0) {
            var curRow = tbody.children[moveNr];
            curRow.setAttribute('id', 'cur-move');

            var topCoord = curRow.offsetTop;
            var bottomCoord = curRow.offsetTop + curRow.offsetHeight - self.moveTable.offsetHeight;
            if (self.moveTable.scrollTop < bottomCoord) {
                self.moveTable.scrollTop = bottomCoord;
            } else if (self.moveTable.scrollTop > topCoord) {
                self.moveTable.scrollTop = topCoord;
            }
        } else {
            self.moveTable.scrollTop = 0;
        }
    };
    this.updateSelectedRow(0);

    this.moveTable.setAttribute('tabindex', '1');
    this.moveTable.onkeydown = function (event) {
        if (event.keyCode == 38 || event.keyCode == 37) {
            self.less.setMove(self.less.getMove() - 1);
        } else if (event.keyCode == 40 || event.keyCode == 39) {
            self.less.setMove(self.less.getMove() + 1);
        } else {
            return;
        }
        event.preventDefault();
        event.stopImmediatePropagation();
    };

    table.appendChild(tbody);
    this.moveTable.appendChild(table);
};


LessController.prototype._createStone = function (player) {
    var tsize = this.stoneSize + this.stonePadding;
    // additional padding around the circle to make a smoother transition
    var canvas = document.createElement('canvas');
    canvas.width = canvas.height = tsize;
    canvas.player = player;

    var context = canvas.getContext('2d');
    context.beginPath();
    context.arc(tsize / 2, tsize / 2, this.stoneSize / 2, 0, 2 * Math.PI, false);
    context.fillStyle = this.colorStone[player];
    context.fill();
    context.lineWidth = 1;
    context.strokeStyle = 'rgba(0,0,0,0.75)';
    context.stroke();
    return canvas;
};

LessController.prototype.updateStones = function (board) {
    for (var row = 0; row < board.length; row++) {
        for (var col = 0; col < board[row].length; col++) {
            var container = this._getCell(row, col);

            var curPlayer = board[row][col];
            var lstPlayer = container.firstChild ? container.firstChild.player : -1;
            if (curPlayer != lstPlayer) {
                while (container.firstChild) container.removeChild(container.firstChild);
                if (curPlayer != -1) {
                    container.appendChild(this._createStone(curPlayer));
                }
            }
        }
    }
};

// required functions of 'interface':
LessController.prototype.setWalls = function (walls) {
    this.createTable();
    this.buildWalls(walls);
};

LessController.prototype.setMoves = function (moves) {
    this.buildMoveTable(moves);
};


LessController.prototype.onBoardChanged = function (move, board) {
    if (this.updateSelectedRow) {
        this.updateSelectedRow(move);
    }
    this.updateStones(board);
};
